# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from concurrent.futures import ThreadPoolExecutor, as_completed
from timeit import default_timer
from datetime import datetime
from threading import Event, Timer
from mktxp.cli.config.config import config_handler
from mktxp.cli.config.config import MKTXPConfigKeys

class CollectorHandler:
    ''' MKTXP Collectors Handler
    '''

    def __init__(self, entries_handler, collector_registry):
        self.entries_handler = entries_handler
        self.collector_registry = collector_registry
        self.last_collect_timestamp = 0


    def collect_sync(self):
        """
        Collect the metrics of all router entries defined in the current users configuration synchronously.
        This function iterates over each router entry one-by-one.
        Thus, the total runtime of this function scales linearly with the number of registered routers.
        """
        for router_entry in self.entries_handler.router_entries:
            if not router_entry.is_ready():
                # let's pick up on things in the next run
                continue

            for collector_ID, collect_func in self.collector_registry.registered_collectors.items():
                start = default_timer()
                yield from collect_func(router_entry)
                router_entry.time_spent[collector_ID] += default_timer() - start
            router_entry.is_done()

    def collect_router_entry_async(self, router_entry, scrape_timeout_event, total_scrape_timeout_event):
        results = []
        for collector_ID, collect_func in self.collector_registry.registered_collectors.items():
            if scrape_timeout_event.is_set():
                print(f'Hit timeout while scraping router entry: {router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME]}')
                break

            if total_scrape_timeout_event.is_set():
                print(f'Hit overall timeout while scraping router entry: {router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME]}')
                break

            start = default_timer()
            result = list(collect_func(router_entry))
            results += result
            router_entry.time_spent[collector_ID] += default_timer() - start        
        router_entry.is_done()
        return results


    def collect_async(self, max_worker_threads=5):
        """
        Collect the metrics of all router entries defined in the current users configuration in parallel.
        This function iterates over multiple routers in parallel (depending on the value of max_worker_threads).
        Thus, the total runtime scales sub linearly (number_of_routers / max_worker_threads).
        """

        def timeout(timeout_event):
            timeout_event.set()

        # overall scrape duration 
        total_scrape_timeout_event = Event()
        total_scrape_timer = Timer(config_handler.system_entry.total_max_scrape_duration, timeout, args=(total_scrape_timeout_event,))
        total_scrape_timer.start()

        with ThreadPoolExecutor(max_workers=max_worker_threads) as executor:
            futures = {}

            for router_entry in self.entries_handler.router_entries:
                if total_scrape_timeout_event.is_set():
                    print(f'Hit overall timeout while scraping router entry: {router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME]}')
                    break

                if not router_entry.is_ready():
                    # let's pick up on things in the next run
                    continue
                
                # Duration of individual scrapes
                scrape_timeout_event = Event()
                scrape_timer = Timer(config_handler.system_entry.max_scrape_duration, timeout, args=(scrape_timeout_event,))
                scrape_timer.start()

                futures[executor.submit(self.collect_router_entry_async, router_entry, scrape_timeout_event, total_scrape_timeout_event)] = scrape_timer

            for future in as_completed(futures):
                # cancel unused timers for scrapes finished regularly (within set duration)
                futures[future].cancel()
                yield from future.result()
            
        # in case collection finished without timeouts, cancel the overall scrape duration timer
        total_scrape_timer.cancel()


    def collect(self):
        if not self._valid_collect_interval():
            return

        # bandwidth collector
        yield from self.collector_registry.bandwidthCollector.collect()

        # all other collectors
        # Check whether to run in parallel by looking at the mktxp system configuration
        parallel = config_handler.system_entry.fetch_routers_in_parallel
        max_worker_threads = config_handler.system_entry.max_worker_threads
        if parallel:
            yield from self.collect_async(max_worker_threads=max_worker_threads)
        else:
            yield from self.collect_sync()


    def _valid_collect_interval(self):
        now = datetime.now().timestamp()
        diff = now - self.last_collect_timestamp
        if diff < config_handler.system_entry.minimal_collect_interval:
            if config_handler.system_entry.verbose_mode:
                print(f'An attempt to collect metrics within minimal metrics collection interval: {diff} < {config_handler.system_entry.minimal_collect_interval}')
                print('deferring..')
            return False

        self.last_collect_timestamp = now       
        return True











# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


import socket
import speedtest
from datetime import datetime
from multiprocessing import Pool, get_context
from mktxp.cli.config.config import config_handler
from mktxp.collector.base_collector import BaseCollector


result_list = [{'download': 0, 'upload': 0, 'ping': 0}]
def get_result(bandwidth_dict):
    result_list[0] = bandwidth_dict


class BandwidthCollector(BaseCollector):
    ''' MKTXP collector
    '''    
    def __init__(self):
        self.pool = None
        self.last_call_timestamp = 0        
    
    def collect(self):
        if not config_handler.system_entry.bandwidth:
            return

        if self.pool is None:
            self.pool = get_context("spawn").Pool()

        if result_list:      
            result_dict = result_list[0]
            bandwidth_records = [{'direction': key, 'bandwidth': str(result_dict[key])} for key in ('download', 'upload')]     
            bandwidth_metrics = BaseCollector.gauge_collector('internet_bandwidth', 'Internet bandwidth in bits per second', 
                                                                            bandwidth_records, 'bandwidth', ['direction'], add_id_labels = False)
            yield bandwidth_metrics

            latency_records = [{'latency': str(result_dict['ping'])}]
            latency_metrics = BaseCollector.gauge_collector('internet_latency', 'Internet latency in milliseconds', 
                                                                            latency_records, 'latency', [], add_id_labels = False)
            yield latency_metrics

        ts =  datetime.now().timestamp()       
        if (ts - self.last_call_timestamp) > config_handler.system_entry.bandwidth_test_interval:
            self.pool.apply_async(BandwidthCollector.bandwidth_worker, callback=get_result)            
            self.last_call_timestamp = ts

    def __del__(self):
        if self.pool is not None:
                self.pool.close()
                self.pool.join()

    @staticmethod
    def bandwidth_worker():
        if BandwidthCollector.inet_connected():
            bandwidth_test = speedtest.Speedtest()
            bandwidth_test.get_best_server()
            bandwidth_test.download()
            bandwidth_test.upload()
            return bandwidth_test.results.dict()
        else:
            return {'download': 0, 'upload': 0, 'ping': 0}

    @staticmethod
    def inet_connected(host="8.8.8.8", port=53, timeout=3):
        try:
            socket.setdefaulttimeout(timeout)
            socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
            return True
        except socket.error as exc:
            return False




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.capsman_ds import CapsmanRegistrationsMetricsDataSource

class CapsmanOutput:
    ''' CAPsMAN CLI Output
    '''    
    @staticmethod
    def clients_summary(router_entry):
        registration_labels = ['interface', 'ssid', 'mac_address', 'rx_signal', 'uptime', 'tx_rate', 'rx_rate']
        registration_records = CapsmanRegistrationsMetricsDataSource.metric_records(router_entry, metric_labels = registration_labels, add_router_id = False)
        if not registration_records:
            print('No CAPsMAN registration records')
            return 

        # translate / trim / augment registration records
        dhcp_rt_by_interface = {}
        for registration_record in sorted(registration_records, key = lambda rt_record: rt_record['rx_signal'], reverse=True):
            BaseOutputProcessor.augment_record(router_entry, registration_record)

            interface = registration_record['interface']
            if interface in dhcp_rt_by_interface.keys():
                dhcp_rt_by_interface[interface].append(registration_record)
            else:
                dhcp_rt_by_interface[interface] = [registration_record]         

        output_records = 0
        registration_records = len(registration_records)                
        output_entry = BaseOutputProcessor.OutputCapsmanEntry
        output_table = BaseOutputProcessor.output_table(output_entry)
                
        for key in dhcp_rt_by_interface.keys():
            for record in dhcp_rt_by_interface[key]:
                output_table.add_row(output_entry(**record))
                output_records += 1
            if output_records < registration_records:
                output_table.add_row(output_entry())

        print (output_table.draw())

        for server in dhcp_rt_by_interface.keys():
            print(f'{server} clients: {len(dhcp_rt_by_interface[server])}')
        print(f'Total connected CAPsMAN clients: {output_records}', '\n')




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from mktxp.datasource.base_ds import BaseDSProcessor
from mktxp.datasource.system_resource_ds import SystemResourceMetricsDataSource
from mktxp.utils.utils import routerOS7_version


class RoutingStatsMetricsDataSource:
    ''' Routing Stats data provider
    '''             
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None, translation_table = None):
        if metric_labels is None:
            metric_labels = []                
        try:
            routing_stats = '/routing/stats/process'

            # legacy 6.x versions are untested
            ver = SystemResourceMetricsDataSource.os_version(router_entry)
            if not routerOS7_version(ver):
                raise Exception("Routing stats for legacy 6.x versions are not supported at the moment")

            routing_stats_records = router_entry.api_connection.router_api().get_resource(routing_stats).get()
            return BaseDSProcessor.trimmed_records(router_entry, router_records = routing_stats_records, metric_labels = metric_labels, translation_table = translation_table)
        except Exception as exc:
            print(f'Error getting routing stats sessions info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.certificate_ds import CertificateMetricsDataSource
from datetime import datetime

class CertificateCollector(BaseCollector):
    '''Certificate collector'''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.certificate:
            return

        certificate_labels = ['name', 'digest_algorithm', 'key_type', 'country', 'state', 'locality', 'organization', 
                                'common_name', 'key_size', 'subject_alt_name', 'days_valid', 'trusted', 'key_usage', 
                                'ca', 'serial_number', 'key_usage', 'ca', 'serial_number', 'fingerprint', 'akid', 'skid', 
                                'invalid_before', 'invalid_after', 'expires_after']

        translation_table = {
        }

        certificate_records = CertificateMetricsDataSource.metric_records(router_entry, translation_table=translation_table, metric_labels = certificate_labels)   
        if isinstance(certificate_records, list): 
            for record in certificate_records:
                # Convert invalid_after time to epoch time
                record['invalid_after_epoch'] = int(datetime.strptime(record['invalid_after'], "%Y-%m-%d %H:%M:%S").timestamp())
        if certificate_records:
            yield BaseCollector.gauge_collector('certificate_expiration_timestamp_seconds', 'The number of seconds before expiration time the certificate should renew.', certificate_records, 'invalid_after_epoch', certificate_labels)



# coding=utf8
# Copyright (c) 2020 Arseniy Kuznetsov
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor


class NeighborDataSource:
    def metric_records(router_entry, metric_labels, ipv6=False):    
        
        metric_labels = metric_labels or []                        
        router_records = []        
        
        try:
            if ipv6:
                router_records = router_entry.api_connection.router_api().get_resource(f'/ipv6/neighbor').get(status='reachable')
            else:
                router_records = router_entry.api_connection.router_api().get_resource(f'/ip/neighbor').get()
            return BaseDSProcessor.trimmed_records(router_entry, router_records=router_records, metric_labels=metric_labels)
        except Exception as exc:
            print(f'Error getting Neighbors info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


import ssl
import socket
import collections
from datetime import datetime
from mktxp.cli.config.config import config_handler
import functools
import yaml

# Fix UTF-8 decode error
# See: https://github.com/akpw/mktxp/issues/47
# The RouterOS-api implicitly assumes that the API response is UTF-8 encoded.
# But Mikrotik uses latin-1.
# Because the upstream dependency is currently abandoned, this is a quick hack to solve the issue

MIKROTIK_ENCODING = 'latin-1'
import routeros_api.api_structure

def _decode_bytes(bytes_to_decode):
    try:
        return bytes_to_decode.decode('utf-8')
    except UnicodeDecodeError:
        return bytes_to_decode.decode(MIKROTIK_ENCODING)

routeros_api.api_structure.StringField.get_python_value = lambda _, bytes:  _decode_bytes(bytes)
routeros_api.api_structure.default_structure = collections.defaultdict(routeros_api.api_structure.StringField)

from routeros_api import RouterOsApiPool

class RouterAPIConnectionError(Exception):
    pass

def check_connected(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.is_connected():
            raise RouterAPIConnectionError(f'No network connection to router: {self.router_name}@{self.config_entry.hostname}')
        else:
            return func(self, *args, **kwargs)
    return wrapper

class RouterAPIConnection:
    ''' Base wrapper interface for the routeros_api library
    '''            
    def __init__(self, router_name, config_entry):
        self.router_name = router_name
        self.config_entry = config_entry
        self.last_failure_timestamp = self.successive_failure_count = 0
        
        ctx = None
        if self.config_entry.use_ssl:
            ctx = ssl.create_default_context(cafile=self.config_entry.ssl_ca_file if self.config_entry.ssl_ca_file else None)
            if self.config_entry.no_ssl_certificate:
                ctx.check_hostname = False
                ctx.verify_mode = ssl.CERT_NONE
            elif self.config_entry.ssl_ca_file:
                ctx.load_verify_locations(self.config_entry.ssl_ca_file)

        username = self.config_entry.username
        password = self.config_entry.password
        if self.config_entry.credentials_file:
            with open(self.config_entry.credentials_file, 'r') as file:
                try:
                    credentials = yaml.safe_load(file)
                    if not isinstance(credentials, dict):
                        raise yaml.YAMLError("Credentials file is not a valid key-value map.")
                except yaml.YAMLError as e:
                    print(f"Error parsing credentials file: {e}\nCheck that it is valid YAML (e.g., 'username: user').")
                    exit(1)                    
                username = credentials.get('username', username)
                password = credentials.get('password', password)

        self.connection = RouterOsApiPool(
                host = self.config_entry.hostname,
                username = username,
                password = password,
                port = self.config_entry.port,
                plaintext_login = True,
                use_ssl = self.config_entry.use_ssl,
                ssl_verify = self.config_entry.ssl_certificate_verify,
                ssl_context = ctx)
        
        self.connection.socket_timeout = config_handler.system_entry.socket_timeout
        self.api = None
    
    def is_connected(self):
        if self.connection and self.connection.connected and self.api:
            return True
        return False

    def connect(self):
        connect_time = datetime.now()
        if self.is_connected() or self._in_connect_timeout(connect_time.timestamp()):
            return
        try:
            print(f'Connecting to router {self.router_name}@{self.config_entry.hostname}')
            self.connection.plaintext_login = self.config_entry.plaintext_login
            self.api = self.connection.get_api()
            self._set_connect_state(success = True, connect_time = connect_time)
        except (socket.error, socket.timeout, Exception) as exc:
            self._set_connect_state(success = False, connect_time = connect_time, exc = exc)
            raise RouterAPIConnectionError(f'Failed attemp to establish network connection to router: {self.router_name}@{self.config_entry.hostname}')

    def disconnect(self):
        if self.is_connected():
            self.connection.disconnect()
            self.api = None

    @check_connected
    def router_api(self):
        return self.api

    def _in_connect_timeout(self, connect_timestamp):
        connect_delay = self._connect_delay()
        if (connect_timestamp - self.last_failure_timestamp) < connect_delay:
            if config_handler.system_entry.verbose_mode: 
                print(f'{self.router_name}@{self.config_entry.hostname}: in connect timeout, {int(connect_delay - (connect_timestamp - self.last_failure_timestamp))}secs remaining')
                print(f'Successive failure count: {self.successive_failure_count}')
            return True
        if config_handler.system_entry.verbose_mode: 
            print(f'{self.router_name}@{self.config_entry.hostname}: OK to connect')
            if self.last_failure_timestamp > 0:
                print(f'Seconds since last failure: {connect_timestamp - self.last_failure_timestamp}')
                print(f'Prior successive failure count: {self.successive_failure_count}')
        return False

    def _connect_delay(self):
        mktxp_entry = config_handler.system_entry
        connect_delay = (1 + self.successive_failure_count / mktxp_entry.delay_inc_div) * mktxp_entry.initial_delay_on_failure
        return connect_delay if connect_delay < mktxp_entry.max_delay_on_failure else mktxp_entry.max_delay_on_failure


    def _set_connect_state(self, success = False, connect_time = datetime.now(), exc = None):
        if success:
            self.last_failure_timestamp = 0
            self.successive_failure_count = 0
            print(f'{connect_time.strftime("%Y-%m-%d %H:%M:%S")} Connection to router {self.router_name}@{self.config_entry.hostname} has been established')
        else:
            self.api = None
            self.successive_failure_count += 1
            self.last_failure_timestamp = connect_time.timestamp() 
            print(f'{connect_time.strftime("%Y-%m-%d %H:%M:%S")} Connection to router {self.router_name}@{self.config_entry.hostname} has failed: {exc}')



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from mktxp.cli.config.config import config_handler
from mktxp.flow.router_entry import RouterEntry
from mktxp.flow.router_connection import RouterAPIConnectionError

class RouterEntriesHandler:
    ''' Handles RouterOS entries defined in MKTXP config 
    '''         
    def __init__(self):
        self._router_entries = {}            
        for router_name in config_handler.registered_entries():            
            router_entry = RouterEntry(router_name)
            RouterEntriesHandler._set_child_entries(router_entry)
            self._router_entries[router_name] = router_entry

    @property
    def router_entries(self):
        return (entry for key, entry in  self._router_entries.items() if entry.config_entry.enabled) \
                                                                                if self._router_entries else None   

    @staticmethod
    def router_entry(entry_name, enabled_only = False):
        ''' A static router entry initialiser
        '''
        config_entry = config_handler.config_entry(entry_name)
        if enabled_only and not config_entry.enabled:
            return None

        router_entry = RouterEntry(entry_name)
        RouterEntriesHandler._set_child_entries(router_entry)
        try:
            router_entry.connect()
        except RouterAPIConnectionError as exc:
            print (f'{exc}')
        return router_entry

    @staticmethod
    def _set_child_entries(router_entry):
        if router_entry.config_entry.remote_dhcp_entry and config_handler.registered_entry(router_entry.config_entry.remote_dhcp_entry):
            router_entry.dhcp_entry = RouterEntry(router_entry.config_entry.remote_dhcp_entry)        

        if router_entry.config_entry.remote_capsman_entry and config_handler.registered_entry(router_entry.config_entry.remote_capsman_entry):
            router_entry.capsman_entry = RouterEntry(router_entry.config_entry.remote_capsman_entry)       




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.interface_ds import InterfaceTrafficMetricsDataSource


class InterfaceCollector(BaseCollector):
    ''' Router Interface Metrics collector
    '''        
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.interface:
            return

        interface_traffic_labels = ['disabled', 'name', 'comment', 'rx_byte', 'tx_byte', 'rx_packet', 'tx_packet',
                                    'rx_error', 'tx_error', 'rx_drop', 'tx_drop', 'link_downs', 'running', 'type']
        interface_traffic_translation_table = {
            'running': lambda value: '1' if value == 'true' else '0',
            'disabled': lambda value: '1' if value == 'true' else '0'
        }

        interface_traffic_records = InterfaceTrafficMetricsDataSource.metric_records(
            router_entry,
            metric_labels=interface_traffic_labels,
            translation_table=interface_traffic_translation_table,
        )

        if not interface_traffic_records:
            return

        yield BaseCollector.info_collector(
            'interface_comment',
            'The interface comment',
            interface_traffic_records,
            metric_labels=['name', 'comment']
        )

        yield BaseCollector.info_collector(
            'interface_type',
            'Interface type like ether, vrrp, eoip, gre-tunnel, ...',
            interface_traffic_records,
            metric_labels=['name', 'type']
        )

        yield BaseCollector.gauge_collector(
            'interface_running',
            'Current running status of the interface',
            interface_traffic_records,
            metric_key='running',
            metric_labels=['name']
        )

        yield BaseCollector.gauge_collector(
            'interface_disabled',
            'Current disabled status of the interface',
            interface_traffic_records,
            metric_key='disabled',
            metric_labels=['name']
        )

        yield BaseCollector.counter_collector(
            'interface_rx_byte',
            'Number of received bytes',
            interface_traffic_records,
            'rx_byte',
            ['name']
        )

        yield BaseCollector.counter_collector(
            'interface_tx_byte',
            'Number of transmitted bytes',
            interface_traffic_records,
            'tx_byte',
            ['name']
        )

        yield BaseCollector.counter_collector(
            'interface_rx_packet',
            'Number of packets received',
            interface_traffic_records,
            'rx_packet',
            ['name']
        )

        yield BaseCollector.counter_collector(
            'interface_tx_packet',
            'Number of transmitted packets',
            interface_traffic_records,
            'tx_packet',
            ['name']
        )

        yield BaseCollector.counter_collector(
            'interface_rx_error',
            'Number of packets received with an error',
            interface_traffic_records,
            'rx_error',
            ['name']
        )

        yield BaseCollector.counter_collector(
            'interface_tx_error',
            'Number of packets transmitted with an error',
            interface_traffic_records,
            'tx_error',
            ['name']
        )

        yield BaseCollector.counter_collector(
            'interface_rx_drop',
            'Number of received packets being dropped',
            interface_traffic_records,
            'rx_drop',
            ['name']
        )

        yield BaseCollector.counter_collector(
            'interface_tx_drop',
            'Number of transmitted packets being dropped',
            interface_traffic_records,
            'tx_drop',
            ['name']
        )

        yield BaseCollector.counter_collector(
            'link_downs',
            'Number of times link went down',
            interface_traffic_records,
            'link_downs',
            ['name']
        )



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from functools import lru_cache
import os, sys, shlex, tempfile, shutil, re
import subprocess, hashlib, urllib
import time
from timeit import default_timer
from collections.abc import Iterable
import xml.etree.ElementTree as ET
from contextlib import contextmanager
from multiprocessing import Process, Event
from datetime import timedelta
from packaging.version import parse

''' Utilities / Helpers
'''
@contextmanager
def temp_dir(quiet = True):
    ''' Temp dir context manager
    '''
    tmp_dir = tempfile.mkdtemp()
    try:
        yield tmp_dir
    finally:
        # remove tmp dir
        try:
            shutil.rmtree(tmp_dir)
        except OSError as e:
            if not quiet:
                print ('Error while removing a tmp dir: {}'.format(e.args[0]))

class CmdProcessingError(Exception):
    pass

def run_cmd(cmd, shell = False, quiet = False):
    ''' Runs shell commands in a separate process
    '''
    if not shell:
        cmd = shlex.split(cmd)
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell = shell)
    output = proc.communicate()[0].decode('utf-8')
    if proc.returncode != 0 and not quiet:
        raise CmdProcessingError(output)
    return output

def parse_mkt_uptime(time):
    time_dict = re.match(r'((?P<weeks>\d+)w)?((?P<days>\d+)d)?((?P<hours>\d+)h)?((?P<minutes>\d+)m)?((?P<seconds>\d+)s)?', time).groupdict()
    delta = timedelta(**{key: int(value) for key, value in time_dict.items() if value}).total_seconds() 
    return int(delta) if delta else 0

def str2bool(str_value, default = None):
    if not str_value or not type(str_value) is str:
        return False
    str_value = str_value.lower()
    if str_value in ('y', 'yes', 't', 'true', 'on', 'ok', '1'):
        return True
    elif str_value in ('n', 'no', 'f', 'false', 'off', 'fail', '0'):
        return False
    elif default is not None:
        return default
    else:
        raise ValueError(f'Invalid truth value: {str_value}')
        
class FSHelper:
    ''' File System ops helper
    '''
    @staticmethod
    def full_path(path, check_parent_path = False):
        ''' Full path
        '''
        if path:
            path = os.path.expanduser(path)
            path = os.path.expandvars(path)
            path = os.path.abspath(path)
            path = os.path.realpath(path)

        # for files, check that the parent dir exists
        if check_parent_path:
            if not os.access(os.path.dirname(path), os.W_OK):
                print('Non-valid folder path:\n\t "{}"'.format(os.path.dirname(path)))
                sys.exit(1)

        return path if path else None

    @staticmethod
    def mountpoint(path):
        ''' The mount point portion of a path
        '''
        path = FSHelper.full_path(path)
        while path != os.path.sep:
            if os.path.ismount(path):
                return path
            path = os.path.realpath(os.path.join(path, os.pardir))
        return path if path != os.path.sep else None

    @staticmethod
    def move_FS_entry(orig_path, target_path,
                      check_unique = True,
                      quiet = False, stop = False):
        ''' Moves FS entry
        '''
        succeeded = False
        try:
            if check_unique and os.path.exists(target_path):
                raise OSError('\nTarget path entry already exists')
            shutil.move(orig_path, target_path)
            succeeded = True
        except OSError as e:
            if not quiet:
                print(str(e))
                print('Failed to move entry:\n\t{0}\n\t{1}'.format(orig_path, target_path))
                print('Exiting...') if stop else print('Skipping...')
            if stop:
                sys.exit(1)
        return succeeded

    @staticmethod
    def file_md5(fpath, block_size=0, hex=False):
        ''' Calculates MD5 hash for a file at fpath
        '''
        md5 = hashlib.md5()
        if block_size == 0:
            block_size = 128 * md5.block_size
        with open(fpath,'rb') as f:
            for chunk in iter(lambda: f.read(block_size), b''):
                md5.update(chunk)
        return md5.hexdigest() if hex else md5.digest()


class UniqueDirNamesChecker:
    ''' Unique file names Helper
    '''
    def __init__(self, src_dir, unique_fnames = None):
        self._uname_gen = unique_fnames() if unique_fnames else self.unique_fnames()

        # init the generator function with file names from given source directory
        src_dir = FSHelper.full_path(src_dir)
        fnames = [fname for fname in os.listdir(src_dir)]

        for fname in fnames:
            next(self._uname_gen)
            self._uname_gen.send(fname)

    def unique_name(self, fname):
        ''' Returns unique file name
        '''
        next(self._uname_gen)
        return self._uname_gen.send(fname)

    @staticmethod
    def unique_fnames():
        ''' default unique file names generator method,
            via appending a simple numbering pattern
        '''
        unique_names = {}
        while True:
            fname = yield
            while True:
                if fname in unique_names:
                    unique_names[fname] += 1
                    name_base, name_ext = os.path.splitext(fname)
                    fname = '{0}_{1}{2}'.format(name_base, unique_names[fname], name_ext)
                else:
                    unique_names[fname] = 0
                    yield fname
                    break


class UniquePartialMatchList(list):
    ''' Enables matching elements by unique "shortcuts"
        e.g:
            >> 'Another' in UniquePartialMatchList(['A long string', 'Another longs string'])
            >> True
            >>'long' in UniquePartialMatchList(['A long string', 'Another longs string'])
            >> False
            >> l.find('Another')
            >> 'Another longs string'
    '''
    def _matched_items(self, partialMatch):
        ''' Generator expression of <matched items>, where <matched item> is
            a tuple of (<matched_element>, <is_exact_match>)
        '''
        def _contains_or_equal(item):
            if isinstance(item, Iterable):
                return (partialMatch in item)
            else:
                return (partialMatch == item)
        return ((item, (partialMatch == item)) for item in self if _contains_or_equal(item))

    def find(self, partialMatch):
        ''' Returns the element in which <partialMatch> can be found
            <partialMatch> is found if it either:
                equals to an element or is contained by exactly one element
        '''
        matched_cnt, unique_match = 0, None
        matched_items = self._matched_items(partialMatch)
        for match, exact_match in matched_items:
            if exact_match:
                # found exact match
                return match
            else:
                # found a partial match
                if not unique_match:
                    unique_match = match
                matched_cnt += 1
        return unique_match if matched_cnt == 1 else None

    def __contains__(self, partialMatch):
        ''' Check if <partialMatch> is contained by an element in the list,
            where <contained> is defined either as:
                either "equals to element" or "contained by exactly one element"
        '''
        return True if self.find(partialMatch) else False


class RepeatableTimer:
    def __init__(self, interval, func, args=[], kwargs={}, process_name = None, repeatable = True, restartable = False):
        self.process_name = process_name
        self.interval = interval
        self.restartable = restartable
        
        self.func = func
        self.args = args
        self.kwargs = kwargs
        
        self.finished = Event()
        self.run_once = Event()
        if not repeatable:
            self.run_once.set()            
        self.process = Process(name = self.process_name, target=self._execute)

    def start(self):
        if self.restartable:
            self.finished.clear()
            self.process = Process(name = self.process_name, target=self._execute, daemon=True)
        self.process.start()

    def stop(self):
        self.finished.set()
        if self.process.is_alive:
            self.process.join()

    def _execute(self):
        while True:
            self.func(*self.args, **self.kwargs)
            if self.finished.is_set() or self.run_once.is_set():
                break
            self.finished.wait(self.interval)     

class Benchmark:
    def __enter__(self):
        self.start = default_timer()
        return self

    def __exit__(self, *args):
        self.time = default_timer() - self.start


# mapping of channels to RSS feeds
CHANNEL_RSS_FEED_MAPPING = {
    'development': 'https://mikrotik.com/development.rss',
    'long-term': 'https://mikrotik.com/bugfix.rss',
    'stable': 'https://mikrotik.com/current.rss',
    'testing': 'https://mikrotik.com/candidate.rss',
}


def get_ttl_hash(seconds=3600):
    """Return the same value withing `seconds` time period"""
    return round(time.time() / seconds)


@lru_cache(maxsize=5)
def get_available_updates(channel, ttl_hash=get_ttl_hash()):
    """Check the RSS feed for available updates for a given update channel.
    This method fetches the RSS feed and returns all version from the parsed XML.
    Version numbers are parsed into version.Version instances (part of setuptools)."""
    del ttl_hash
    rss_feed = CHANNEL_RSS_FEED_MAPPING[channel]

    print(f'Fetching available ROS releases from {rss_feed}')
    versions = []
    with urllib.request.urlopen(rss_feed) as response:
        result = response.read()
        root = ET.fromstring(result)
        channel = root[0]

        for child in channel:
            # iterate over all updates
            if child.tag == 'item':
                title = child[0]
                # extract and parse the version number from title
                version_text = re.findall(r'[\d+\.]+', title.text)[0]
                version_number = parse(version_text)
                versions.append(version_number)
    return versions


def parse_ros_version(string):
    """Parse the version returned from the /system/resource command.
    Returns a tuple: (<version>, <channel>).

    >>> parse_ros_version('1.2.3 (stable)')
    1.2.3, stable

    >>> parse_ros_version('7.14.3 (long-term)')
    7.14.3, long-term    
    """

    match = re.findall(r'([\d\.]+).*?\(([\w-]+)\)', string)
    if match:
        version, channel = match[0]
        return parse(version), channel
    else:
        # Fallback for versions without parentheses or non-standard format
        version_match = re.findall(r'([\d\.]+)', string)
        if version_match:
            return parse(version_match[0]), 'stable'  # Default to stable if no channel found
        else:
            raise ValueError(f'Unable to parse RouterOS version from: {string}')


def builtin_wifi_capsman_version(version):
    """Try to check if the version is Wifi version of RouterOS (>= 7.13).
    Returns a boolean"""
    try:
        cur_version, _ = parse_ros_version(version)
        if cur_version >= parse('7.13'):
            return True
    except Exception as err:
        print(f'could not get current RouterOS version, because: {err}')
    return False

def routerOS7_version(version):
    try:
        cur_version, _ = parse_ros_version(version)
        if cur_version >= parse('7.0'):
            return True
    except Exception as err:
        print(f'could not get current RouterOS version, because: {err}')
    return False

def check_for_updates(cur_version):
    """Try to check if there is a newer version available.
    If anything goes wrong, it returns the same version.
    Returns a tuple: (<current version>, <newest version>)"""
    error = False
    try:
        cur_version, channel = parse_ros_version(cur_version)
        available_versions = get_available_updates(channel)
        newest_version = sorted(available_versions)[-1]
    except KeyError:
        print(f'unknown update channel {channel}')
        error = True
    except urllib.error.HTTPError as err:
        print(f'update feed returned: {err}')
        error = True
    except Exception as err:
        print(f'could not check for updates, because: {err}')
        error = True

    if error:
        return cur_version, cur_version

    return cur_version, newest_version



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.cli.config.config import MKTXPConfigKeys
from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.pool_ds import PoolMetricsDataSource, PoolUsedMetricsDataSource


class PoolCollector(BaseCollector):
    ''' IP Pool Metrics collector
    '''    
    @staticmethod
    def collect(router_entry):
        # ~*~*~*~*~*~ IPv4 ~*~*~*~*~*~
        if router_entry.config_entry.pool:
            yield from PoolCollector._process_ip_stack(router_entry)

        # ~*~*~*~*~*~ IPv6 ~*~*~*~*~*~
        if router_entry.config_entry.ipv6_pool:
            yield from PoolCollector._process_ip_stack(router_entry, ipv6=True)

    # helpers
    @staticmethod
    def _process_ip_stack(router_entry, ipv6=False):
        ip_stack = 'ipv6' if ipv6 else 'ipv4'

        # initialize all pool counts, including those currently not used
        pool_records = PoolMetricsDataSource.metric_records(router_entry, metric_labels = ['name'], ipv6=ipv6)   
        if pool_records:
            pool_used_labels = ['pool']
            pool_used_counts = {pool_record['name']: 0 for pool_record in pool_records}

            # for pools in usage, calculate the current numbers
            pool_used_records = PoolUsedMetricsDataSource.metric_records(router_entry, metric_labels = pool_used_labels, ipv6=ipv6)   
            for pool_used_record in pool_used_records:
                pool_used_counts[pool_used_record['pool']] = pool_used_counts.get(pool_used_record['pool'], 0) + 1

           # compile used-per-pool records
            used_per_pool_records = [{ MKTXPConfigKeys.ROUTERBOARD_NAME: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME],
                                       MKTXPConfigKeys.ROUTERBOARD_ADDRESS: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_ADDRESS],
                                       'pool': key, 'count': value} for key, value in pool_used_counts.items()]
            
            # yield used-per-pool metrics
            used_per_pool_metrics = BaseCollector.gauge_collector(f'ip_pool_used{"_ipv6" if ipv6 else ""}', f'Number of used addresses per IP pool ({ip_stack.upper()})', used_per_pool_records, 'count', ['pool'])
            yield used_per_pool_metrics



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.interface_ds import InterfaceMonitorMetricsDataSource
from mktxp.utils.utils import parse_mkt_uptime


class LTECollector(BaseCollector):
    ''' LTE Metrics collector
    '''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.lte:
            return

        monitor_labels = ['pin_status', 'registration_status', 'functionality', 'current_operator', 'access_technology', 'session_uptime', 'subscriber_number', 'rsrp', 'rsrq']
        translation_table = {
                'pin_status': lambda value: '1' if value=='ok' else '0',
                'registration_status': lambda value: '1' if value=='registered' else '0',
                'session_uptime': lambda value:  parse_mkt_uptime(value)
                }
        monitor_records = InterfaceMonitorMetricsDataSource.metric_records(router_entry,
                                                                           translation_table=translation_table,
                                                                           kind = 'lte',
                                                                           running_only = False)
        if monitor_records:
            yield BaseCollector.gauge_collector('lte_pin_status', 'Pin status', monitor_records, 'pin_status', [])
            yield BaseCollector.gauge_collector('lte_registration_status', 'Registration status', monitor_records, 'registration_status', [])
            yield BaseCollector.info_collector('lte_current_operator', 'LTE operator', monitor_records, ['current_operator', 'access_technology', 'functionality', 'subscriber_number'])
            yield BaseCollector.gauge_collector('lte_session_uptime', 'LTE session uptime', monitor_records, 'session_uptime', [])

            # specific labels
            rsrp_records = [monitor_record for monitor_record in monitor_records if monitor_record.get('rsrp')]
            rsrq_records = [monitor_record for monitor_record in monitor_records if monitor_record.get('rsrq')]
            sinr_records = [monitor_record for monitor_record in monitor_records if monitor_record.get('sinr')]
            rssi_records = [monitor_record for monitor_record in monitor_records if monitor_record.get('rssi')]

            if rsrp_records:
                yield BaseCollector.gauge_collector('lte_rsrp', 'LTE Reference Signal Received Qualityrsrp value', rsrp_records, 'rsrp', [])
            if rsrq_records:
                yield BaseCollector.gauge_collector('lte_rsrq', 'LTE Referenz Signal Received Power value', rsrq_records, 'rsrq', [])
            if sinr_records:
                yield BaseCollector.gauge_collector('lte_sinr', 'LTE signal to noise ratio value', sinr_records, 'sinr', [])
            if rssi_records:
                yield BaseCollector.gauge_collector('lte_rssi', 'Received Signal Strength Indicator', rssi_records, 'rssi', [])



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.datasource.kid_control_device_ds import KidDeviceMetricsDataSource


class KidDeviceCollector(BaseCollector):
    """ Kid-control device Metrics collector
    """

    @staticmethod
    def collect(router_entry):
        if not (router_entry.config_entry.kid_control_assigned or router_entry.config_entry.kid_control_dynamic):
            return

        labels = ['name', 'user', 'mac_address', 'ip_address', 'bytes_down', 'bytes_up', 'rate_up', 
                    'rate_down','bytes_up', 'idle_time','blocked', 'limited', 'inactive', 'disabled']

        translation_table = {
            'rate_up': lambda value: BaseOutputProcessor.parse_rates(value),
            'rate_down': lambda value: BaseOutputProcessor.parse_rates(value),
            'idle_time': lambda value: BaseOutputProcessor.parse_timedelta_seconds(value) if value else 0,
            'blocked': lambda value: '1' if value == 'true' else '0',
            'limited': lambda value: '1' if value == 'true' else '0',
            'inactive': lambda value: '1' if value == 'true' else '0',
            'disabled': lambda value: '1' if value == 'true' else '0'}

        records = KidDeviceMetricsDataSource.metric_records(router_entry, metric_labels=labels, translation_table=translation_table)
        if records:
            # dhcp resolution
            for registration_record in records:
                BaseOutputProcessor.resolve_dhcp(router_entry, registration_record, resolve_address=False)

            info_labels = ['name', 'dhcp_name', 'mac_address', 'user', 'ip_address', 'disabled']
            yield BaseCollector.info_collector('kid_control_device', 'Kid-control device Info', records, info_labels)

            id_labels = ['name', 'dhcp_name', 'mac_address', 'user']
            yield BaseCollector.counter_collector('kid_control_device_bytes_down', 'Number of received bytes', records, 'bytes_down', id_labels)
            yield BaseCollector.counter_collector('kid_control_device_bytes_up', 'Number of transmitted bytes', records, 'bytes_up', id_labels)

            yield BaseCollector.gauge_collector('kid_control_device_rate_up', 'Device rate up', records, 'rate_up', id_labels)        
            yield BaseCollector.gauge_collector('kid_control_device_idle_time', 'Device idle time', records, 'idle_time', id_labels)
            yield BaseCollector.gauge_collector('kid_control_device_rate_down', 'Device rate down', records, 'rate_down', id_labels)



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor


class NetwatchMetricsDataSource:
    ''' Netwatch Metrics data provider
    '''             
    @staticmethod
    def metric_records(router_entry, *, metric_labels=None, translation_table=None):
        if metric_labels is None:
            metric_labels = []                
        try:
            netwatch_records = router_entry.api_connection.router_api().get_resource('/tool/netwatch').get()
            netwatch_records = [entry for entry in netwatch_records if entry.get('disabled', 'false') != 'true']
            if 'name' in metric_labels:
                for netwatch_record in netwatch_records:
                    comment = netwatch_record.get('comment')
                    host = netwatch_record.get('host')        
                    if comment:
                        netwatch_record['name'] = f'{host} ({comment[0:20]})' if not router_entry.config_entry.use_comments_over_names else comment
                    else:
                        netwatch_record['name'] = host

            return BaseDSProcessor.trimmed_records(router_entry, router_records = netwatch_records, translation_table = translation_table, metric_labels = metric_labels)
        except Exception as exc:
            print(f'Error getting Netwatch info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.cli.config.config import MKTXPConfigKeys
from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.route_ds import RouteMetricsDataSource
from mktxp.utils.utils import str2bool


class RouteCollector(BaseCollector):
    ''' IP Route Metrics collector
    '''        
    @staticmethod
    def collect(router_entry):
        route_labels = ['connect', 'dynamic', 'static', 'bgp', 'ospf']

        # ~*~*~*~*~*~ IPv4 ~*~*~*~*~*~        
        if router_entry.config_entry.route:            
            route_records = RouteMetricsDataSource.metric_records(router_entry, metric_labels = route_labels)   
            if route_records:       
                # compile total routes records
                total_routes = len(route_records)
                total_routes_records = [{ MKTXPConfigKeys.ROUTERBOARD_NAME: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME],
                                          MKTXPConfigKeys.ROUTERBOARD_ADDRESS: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_ADDRESS],
                                          'count': total_routes
                                        }]
                total_routes_metrics = BaseCollector.gauge_collector('routes_total_routes', 'Overall number of routes in RIB', total_routes_records, 'count')
                yield total_routes_metrics


                # init routes per protocol (with 0)
                routes_per_protocol = {route_label: 0 for route_label in route_labels}
                for route_record in route_records:
                    for route_label in route_labels:
                        if str2bool(route_record.get(route_label)):
                            routes_per_protocol[route_label] += 1 

                # compile route-per-protocol records
                route_per_protocol_records = [{ MKTXPConfigKeys.ROUTERBOARD_NAME: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME],
                                                MKTXPConfigKeys.ROUTERBOARD_ADDRESS: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_ADDRESS],
                                                'protocol': key, 'count': value} for key, value in routes_per_protocol.items()]
                
                # yield route-per-protocol metrics
                route_per_protocol_metrics = BaseCollector.gauge_collector('routes_protocol_count', 'Number of routes per protocol in RIB', route_per_protocol_records, 'count', ['protocol'])
                yield route_per_protocol_metrics

        # ~*~*~*~*~*~ IPv6 ~*~*~*~*~*~
        if router_entry.config_entry.ipv6_route:
            # ~*~*~*~*~*~ IPv4 ~*~*~*~*~*~        
            route_records = RouteMetricsDataSource.metric_records(router_entry, metric_labels = route_labels, ipv6 = True)   
            if route_records:       
                # compile total routes records
                total_routes = len(route_records)
                total_routes_records = [{ MKTXPConfigKeys.ROUTERBOARD_NAME: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME],
                                          MKTXPConfigKeys.ROUTERBOARD_ADDRESS: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_ADDRESS],
                                          'count': total_routes
                                        }]
                total_routes_metrics = BaseCollector.gauge_collector('routes_total_routes_ipv6', 'Overall number of routes in RIB (IPv6)', total_routes_records, 'count')
                yield total_routes_metrics


                # init routes per protocol (with 0)
                routes_per_protocol = {route_label: 0 for route_label in route_labels}
                for route_record in route_records:
                    for route_label in route_labels:
                        if route_record.get(route_label):
                            routes_per_protocol[route_label] += 1 

                # compile route-per-protocol records
                route_per_protocol_records = [{ MKTXPConfigKeys.ROUTERBOARD_NAME: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME],
                                                MKTXPConfigKeys.ROUTERBOARD_ADDRESS: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_ADDRESS],
                                                'protocol': key, 'count': value} for key, value in routes_per_protocol.items()]
                
                # yield route-per-protocol metrics
                route_per_protocol_metrics = BaseCollector.gauge_collector('routes_protocol_count_ipv6', 'Number of routes per protocol in RIB (IPv6)', route_per_protocol_records, 'count', ['protocol'])
                yield route_per_protocol_metrics




## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

from setuptools import setup, find_packages
from os import path

# read the README.md contents
pkg_dir = path.abspath(path.dirname(__file__))
with open(path.join(pkg_dir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='mktxp',
    version='1.2.12',

    url='https://github.com/akpw/mktxp',

    author='Arseniy Kuznetsov',
    author_email='k.arseniy@gmail.com',

    long_description=long_description,
    long_description_content_type='text/markdown',

    description=('''
                    Prometheus Exporter for Mikrotik RouterOS devices
                '''),
    license='GNU General Public License v2 (GPLv2)',

    packages=find_packages(exclude=['test*']),

    package_data = {
        '': ['config/*.conf'],
    },

    keywords = 'Mikrotik RouterOS Prometheus Exporter',

    install_requires = ['prometheus-client>=0.9.0', 
                        'RouterOS-api>=0.19.0',
                        'configobj>=5.0.6',
                        'humanize>=3.2.0',
                        'texttable>=1.6.3',
                        'speedtest-cli>=2.1.2',
                        'waitress>=3.0.0',
                        'packaging>=24.0',
                        'pyyaml>=5.1'
                        ],

    test_suite = 'tests',

    extras_require={
        'test': [
            'pytest',
            'pytest-mock',
        ],
    },
    
    entry_points={'console_scripts': [
        'mktxp = mktxp.cli.dispatch:main',
    ]},

    zip_safe=True,

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Customer Service',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: BSD :: FreeBSD',
        'Operating System :: POSIX :: Linux',
        'Topic :: System',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities'
    ]
)






# coding=utf8
# Copyright (c) 2020 Arseniy Kuznetsov
##
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
##
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import os
import sys
import shutil
from collections import namedtuple
from configobj import ConfigObj
from abc import ABCMeta, abstractmethod
import importlib.resources
from mktxp.utils.utils import FSHelper


''' MKTXP conf file handling
'''

class CollectorKeys:
    IDENTITY_COLLECTOR = 'IdentityCollector'
    SYSTEM_RESOURCE_COLLECTOR = 'SystemResourceCollector'
    HEALTH_COLLECTOR = 'HealthCollector'
    PUBLIC_IP_ADDRESS_COLLECTOR = 'PublicIPAddressCollector'
    NEIGHBOR_COLLECTOR = 'NeighborCollector'
    DNS_COLLECTOR = 'DNSCollector'
    PACKAGE_COLLECTOR = 'PackageCollector'
    DHCP_COLLECTOR = 'DHCPCollector'
    POOL_COLLECTOR = 'PoolCollector'
    IP_CONNECTION_COLLECTOR = 'IPConnectionCollector'
    INTERFACE_COLLECTOR = 'InterfaceCollector'
    FIREWALL_COLLECTOR = 'FirewallCollector'
    MONITOR_COLLECTOR = 'MonitorCollector'
    POE_COLLECTOR = 'POECollector'
    NETWATCH_COLLECTOR = 'NetwatchCollector'
    ROUTE_COLLECTOR = 'RouteCollector'
    WLAN_COLLECTOR = 'WLANCollector'
    CAPSMAN_COLLECTOR = 'CapsmanCollector'
    QUEUE_TREE_COLLECTOR = 'QueueTreeCollector'
    QUEUE_SIMPLE_COLLECTOR = 'QueueSimpleCollector'
    KID_CONTROL_DEVICE_COLLECTOR = 'KidControlCollector'
    USER_COLLECTOR = 'UserCollector'
    BFD_COLLECTOR = 'BFDCollector'
    BGP_COLLECTOR = 'BGPCollector'
    ROUTING_STATS_COLLECTOR = 'RoutingStatsCollector'
    EOIP_COLLECTOR = 'EOIPCollector'
    GRE_COLLECTOR = 'GRECollector'
    IPIP_COLLECTOR = 'IPIPCollector'
    IPSEC_COLLECTOR = 'IPSecCollector'
    LTE_COLLECTOR = 'LTECollector'
    SWITCH_PORT_COLLECTOR = 'SwitchPortCollector'
    MKTXP_COLLECTOR = 'MKTXPCollector'
    CERTIFICATE_COLLECTOR = 'CertificateCollector'
    CONTAINER_COLLECTOR = "ContainerCollector"


class MKTXPConfigKeys:
    ''' MKTXP config file keys
    '''
    # Section Keys
    ENABLED_KEY = 'enabled'
    HOST_KEY = 'hostname'
    PORT_KEY = 'port'
    LISTEN_KEY = 'listen'
    USER_KEY = 'username'
    PASSWD_KEY = 'password'
    CREDENTIALS_FILE_KEY = 'credentials_file'

    SSL_KEY = 'use_ssl'
    NO_SSL_CERTIFICATE = 'no_ssl_certificate'
    SSL_CERTIFICATE_VERIFY = 'ssl_certificate_verify'
    SSL_CA_FILE = 'ssl_ca_file'
    PLAINTEXT_LOGIN_KEY = 'plaintext_login'

    FE_HEALTH_KEY = 'health'
    FE_PACKAGE_KEY = 'installed_packages'
    FE_DHCP_KEY = 'dhcp'
    FE_DHCP_LEASE_KEY = 'dhcp_lease'
    FE_IP_CONNECTIONS_KEY = 'connections'
    FE_CONNECTION_STATS_KEY = 'connection_stats'
    FE_INTERFACE_KEY = 'interface'
    
    FE_ROUTE_KEY = 'route'
    FE_DHCP_POOL_KEY = 'pool'
    FE_FIREWALL_KEY = 'firewall'
    FE_NEIGHBOR_KEY = 'neighbor'
    FE_DNS_KEY = 'dns'

    FE_IPV6_ROUTE_KEY = 'ipv6_route'
    FE_IPV6_DHCP_POOL_KEY = 'ipv6_pool'
    FE_IPV6_FIREWALL_KEY = 'ipv6_firewall'
    FE_IPV6_NEIGHBOR_KEY = 'ipv6_neighbor'

    FE_MONITOR_KEY = 'monitor'
    FE_WIRELESS_KEY = 'wireless'
    FE_WIRELESS_CLIENTS_KEY = 'wireless_clients'
    FE_CAPSMAN_KEY = 'capsman'
    FE_CAPSMAN_CLIENTS_KEY = 'capsman_clients'
    FE_POE_KEY = 'poe'
    FE_PUBLIC_IP_KEY = 'public_ip'
    FE_NETWATCH_KEY = 'netwatch'

    FE_EOIP_KEY = 'eoip'
    FE_GRE_KEY = 'gre'
    FE_IPIP_KEY = 'ipip'
    FE_IPSEC_KEY = 'ipsec'
    FE_LTE_KEY = "lte"
    FE_SWITCH_PORT_KEY = "switch_port"

    FE_USER_KEY = 'user'
    FE_QUEUE_KEY = 'queue'
    FE_BFD_KEY = 'bfd'
    FE_BGP_KEY = 'bgp'

    FE_REMOTE_DHCP_ENTRY = 'remote_dhcp_entry'
    FE_REMOTE_CAPSMAN_ENTRY = 'remote_capsman_entry'

    FE_CHECK_FOR_UPDATES = 'check_for_updates'

    FE_KID_CONTROL_DEVICE = 'kid_control_assigned'
    FE_KID_CONTROL_DYNAMIC = 'kid_control_dynamic'

    FE_CONTAINER_KEY = 'container'

    FE_CERTIFICATE_KEY = 'certificate'
    FE_ROUTING_STATS_KEY = 'routing_stats'

    MKTXP_SOCKET_TIMEOUT = 'socket_timeout'
    MKTXP_INITIAL_DELAY = 'initial_delay_on_failure'
    MKTXP_MAX_DELAY = 'max_delay_on_failure'
    MKTXP_INC_DIV = 'delay_inc_div'
    MKTXP_BANDWIDTH_KEY = 'bandwidth'
    MKTXP_BANDWIDTH_TEST_INTERVAL = 'bandwidth_test_interval'
    MKTXP_VERBOSE_MODE = 'verbose_mode'
    MKTXP_MIN_COLLECT_INTERVAL = 'minimal_collect_interval'
    MKTXP_FETCH_IN_PARALLEL = 'fetch_routers_in_parallel'
    MKTXP_MAX_WORKER_THREADS = 'max_worker_threads'
    MKTXP_MAX_SCRAPE_DURATION = 'max_scrape_duration'
    MKTXP_TOTAL_MAX_SCRAPE_DURATION = 'total_max_scrape_duration'
    MKTXP_COMPACT_CONFIG = 'compact_default_conf_values'
    MKTXP_PROMETHEUS_HEADERS_DEDUPLICATION = 'prometheus_headers_deduplication'
    MKTXP_PERSISTENT_ROUTER_CONNECTION_POOL = 'persistent_router_connection_pool'

    # UnRegistered entries placeholder
    NO_ENTRIES_REGISTERED = 'NoEntriesRegistered'

    MKTXP_USE_COMMENTS_OVER_NAMES = 'use_comments_over_names'

    # Base router id labels
    ROUTERBOARD_NAME = 'routerboard_name'
    ROUTERBOARD_ADDRESS = 'routerboard_address'

    # Default values    
    DEFAULT_HOST_KEY = 'localhost'
    DEFAULT_USER_KEY = 'user'
    DEFAULT_PASSWORD_KEY = 'password'
    DEFAULT_CREDENTIALS_FILE_KEY = ""

    DEFAULT_SSL_CA_FILE = ""

    DEFAULT_API_PORT = 8728
    DEFAULT_API_SSL_PORT = 8729
    DEFAULT_FE_REMOTE_DHCP_ENTRY = 'None'
    DEFAULT_FE_REMOTE_CAPSMAN_ENTRY = 'None'
    DEFAULT_MKTXP_PORT = 49090
    DEFAULT_MKTXP_SOCKET_TIMEOUT = 2
    DEFAULT_MKTXP_INITIAL_DELAY = 120
    DEFAULT_MKTXP_MAX_DELAY = 900
    DEFAULT_MKTXP_INC_DIV = 5
    DEFAULT_MKTXP_BANDWIDTH_TEST_INTERVAL = 420
    DEFAULT_MKTXP_MIN_COLLECT_INTERVAL = 5
    DEFAULT_MKTXP_MAX_WORKER_THREADS = 5
    DEFAULT_MKTXP_MAX_SCRAPE_DURATION = 10
    DEFAULT_MKTXP_TOTAL_MAX_SCRAPE_DURATION = 30


    BOOLEAN_KEYS_NO = {ENABLED_KEY, SSL_KEY, NO_SSL_CERTIFICATE, FE_CHECK_FOR_UPDATES, FE_KID_CONTROL_DEVICE, FE_KID_CONTROL_DYNAMIC,
                       SSL_CERTIFICATE_VERIFY, FE_IPV6_ROUTE_KEY, FE_IPV6_DHCP_POOL_KEY, FE_IPV6_FIREWALL_KEY, FE_IPV6_NEIGHBOR_KEY, FE_CONNECTION_STATS_KEY, FE_BFD_KEY, FE_BGP_KEY,
                       FE_EOIP_KEY, FE_GRE_KEY, FE_IPIP_KEY, FE_IPSEC_KEY, FE_LTE_KEY, FE_SWITCH_PORT_KEY, FE_ROUTING_STATS_KEY, FE_CERTIFICATE_KEY, FE_DNS_KEY, FE_CONTAINER_KEY}

    # Feature keys enabled by default
    BOOLEAN_KEYS_YES = {PLAINTEXT_LOGIN_KEY, FE_DHCP_KEY, FE_HEALTH_KEY, FE_PACKAGE_KEY, FE_DHCP_LEASE_KEY, FE_IP_CONNECTIONS_KEY, FE_INTERFACE_KEY, 
                        FE_ROUTE_KEY, FE_DHCP_POOL_KEY, FE_FIREWALL_KEY, FE_NEIGHBOR_KEY, FE_MONITOR_KEY, MKTXP_USE_COMMENTS_OVER_NAMES,
                        FE_WIRELESS_KEY, FE_WIRELESS_CLIENTS_KEY, FE_CAPSMAN_KEY, FE_CAPSMAN_CLIENTS_KEY, FE_POE_KEY,
                        FE_NETWATCH_KEY, FE_PUBLIC_IP_KEY, FE_USER_KEY, FE_QUEUE_KEY}

    SYSTEM_BOOLEAN_KEYS_YES = {MKTXP_PERSISTENT_ROUTER_CONNECTION_POOL}
    SYSTEM_BOOLEAN_KEYS_NO = {MKTXP_BANDWIDTH_KEY, MKTXP_VERBOSE_MODE, MKTXP_FETCH_IN_PARALLEL, MKTXP_COMPACT_CONFIG, MKTXP_PROMETHEUS_HEADERS_DEDUPLICATION}

    STR_KEYS = (HOST_KEY, USER_KEY, PASSWD_KEY, CREDENTIALS_FILE_KEY, SSL_CA_FILE, FE_REMOTE_DHCP_ENTRY, FE_REMOTE_CAPSMAN_ENTRY)
    INT_KEYS =  ()
    MKTXP_INT_KEYS = (PORT_KEY, MKTXP_SOCKET_TIMEOUT, MKTXP_INITIAL_DELAY, MKTXP_MAX_DELAY,
                      MKTXP_INC_DIV, MKTXP_BANDWIDTH_TEST_INTERVAL, MKTXP_MIN_COLLECT_INTERVAL,
                      MKTXP_MAX_WORKER_THREADS, MKTXP_MAX_SCRAPE_DURATION, MKTXP_TOTAL_MAX_SCRAPE_DURATION)

    # MKTXP configs entry names
    DEFAULT_ENTRY_KEY = 'default'
    MKTXP_CONFIG_ENTRY_NAME = 'MKTXP'


class ConfigEntry:
    MKTXPConfigEntry = namedtuple('MKTXPConfigEntry', [MKTXPConfigKeys.ENABLED_KEY, MKTXPConfigKeys.HOST_KEY, MKTXPConfigKeys.PORT_KEY,
                                                       MKTXPConfigKeys.USER_KEY, MKTXPConfigKeys.PASSWD_KEY, MKTXPConfigKeys.CREDENTIALS_FILE_KEY,
                                                       MKTXPConfigKeys.SSL_KEY, MKTXPConfigKeys.NO_SSL_CERTIFICATE, MKTXPConfigKeys.SSL_CERTIFICATE_VERIFY, MKTXPConfigKeys.SSL_CA_FILE, MKTXPConfigKeys.PLAINTEXT_LOGIN_KEY,
                                                       MKTXPConfigKeys.FE_DHCP_KEY, MKTXPConfigKeys.FE_HEALTH_KEY, MKTXPConfigKeys.FE_PACKAGE_KEY, MKTXPConfigKeys.FE_DHCP_LEASE_KEY, MKTXPConfigKeys.FE_INTERFACE_KEY,
                                                       MKTXPConfigKeys.FE_MONITOR_KEY, MKTXPConfigKeys.FE_WIRELESS_KEY, MKTXPConfigKeys.FE_WIRELESS_CLIENTS_KEY,
                                                       MKTXPConfigKeys.FE_IP_CONNECTIONS_KEY, MKTXPConfigKeys.FE_CONNECTION_STATS_KEY, MKTXPConfigKeys.FE_CAPSMAN_KEY, MKTXPConfigKeys.FE_CAPSMAN_CLIENTS_KEY, MKTXPConfigKeys.FE_POE_KEY, 
                                                       MKTXPConfigKeys.FE_NETWATCH_KEY, MKTXPConfigKeys.MKTXP_USE_COMMENTS_OVER_NAMES, MKTXPConfigKeys.FE_PUBLIC_IP_KEY,
                                                       MKTXPConfigKeys.FE_ROUTE_KEY, MKTXPConfigKeys.FE_DHCP_POOL_KEY, MKTXPConfigKeys.FE_FIREWALL_KEY, MKTXPConfigKeys.FE_NEIGHBOR_KEY, MKTXPConfigKeys.FE_DNS_KEY,
                                                       MKTXPConfigKeys.FE_IPV6_ROUTE_KEY, MKTXPConfigKeys.FE_IPV6_DHCP_POOL_KEY, MKTXPConfigKeys.FE_IPV6_FIREWALL_KEY, MKTXPConfigKeys.FE_IPV6_NEIGHBOR_KEY,                                               
                                                       MKTXPConfigKeys.FE_USER_KEY, MKTXPConfigKeys.FE_QUEUE_KEY, MKTXPConfigKeys.FE_REMOTE_DHCP_ENTRY, MKTXPConfigKeys.FE_REMOTE_CAPSMAN_ENTRY, MKTXPConfigKeys.FE_CHECK_FOR_UPDATES, MKTXPConfigKeys.FE_BFD_KEY, MKTXPConfigKeys.FE_BGP_KEY,
                                                       MKTXPConfigKeys.FE_KID_CONTROL_DEVICE, MKTXPConfigKeys.FE_KID_CONTROL_DYNAMIC, MKTXPConfigKeys.FE_EOIP_KEY, MKTXPConfigKeys.FE_GRE_KEY, MKTXPConfigKeys.FE_IPIP_KEY, MKTXPConfigKeys.FE_LTE_KEY, MKTXPConfigKeys.FE_IPSEC_KEY, MKTXPConfigKeys.FE_SWITCH_PORT_KEY,
                                                       MKTXPConfigKeys.FE_ROUTING_STATS_KEY, MKTXPConfigKeys.FE_CERTIFICATE_KEY, MKTXPConfigKeys.FE_CONTAINER_KEY
                                                       ])
    MKTXPSystemEntry = namedtuple('MKTXPSystemEntry', [MKTXPConfigKeys.PORT_KEY, MKTXPConfigKeys.LISTEN_KEY, MKTXPConfigKeys.MKTXP_SOCKET_TIMEOUT,
                                                       MKTXPConfigKeys.MKTXP_INITIAL_DELAY, MKTXPConfigKeys.MKTXP_MAX_DELAY,
                                                       MKTXPConfigKeys.MKTXP_INC_DIV, MKTXPConfigKeys.MKTXP_BANDWIDTH_KEY,
                                                       MKTXPConfigKeys.MKTXP_VERBOSE_MODE, MKTXPConfigKeys.MKTXP_BANDWIDTH_TEST_INTERVAL,
                                                       MKTXPConfigKeys.MKTXP_MIN_COLLECT_INTERVAL, MKTXPConfigKeys.MKTXP_FETCH_IN_PARALLEL,
                                                       MKTXPConfigKeys.MKTXP_MAX_WORKER_THREADS, MKTXPConfigKeys.MKTXP_MAX_SCRAPE_DURATION, 
                                                       MKTXPConfigKeys.MKTXP_TOTAL_MAX_SCRAPE_DURATION, MKTXPConfigKeys.MKTXP_COMPACT_CONFIG, 
                                                       MKTXPConfigKeys.MKTXP_PROMETHEUS_HEADERS_DEDUPLICATION, MKTXPConfigKeys.MKTXP_PERSISTENT_ROUTER_CONNECTION_POOL]) 


class OSConfig(metaclass=ABCMeta):
    ''' OS-related config
    '''
    @staticmethod
    def os_config():
        ''' Factory method
        '''
        if sys.platform == 'linux':
            return LinuxConfig()
        elif sys.platform == 'darwin':
            return OSXConfig()
        elif sys.platform.startswith('freebsd'):
            return FreeBSDConfig()
        else:
            print(f'Non-supported platform: {sys.platform}')
            return None

    @property
    @abstractmethod
    def mktxp_user_dir_path(self):
        pass


class FreeBSDConfig(OSConfig):
    ''' FreeBSD-related config
    '''
    @property
    def mktxp_user_dir_path(self):
        return FSHelper.full_path('~/mktxp')


class OSXConfig(OSConfig):
    ''' OSX-related config
    '''
    @property
    def mktxp_user_dir_path(self):
        return FSHelper.full_path('~/mktxp')


class LinuxConfig(OSConfig):
    ''' Linux-related config
    '''
    @property
    def mktxp_user_dir_path(self):
        return FSHelper.full_path('~/mktxp')


class CustomConfig(OSConfig):
    ''' Custom config
    '''
    def __init__(self, path):
        self._user_dir_path = path

    @property
    def mktxp_user_dir_path(self):
        return FSHelper.full_path(self._user_dir_path)

# Mock system entry to enable running tests
mockSystemEntry = ConfigEntry.MKTXPSystemEntry(
            port=MKTXPConfigKeys.DEFAULT_MKTXP_PORT,
            listen=f'0.0.0.0:{MKTXPConfigKeys.DEFAULT_MKTXP_PORT}',
            socket_timeout=MKTXPConfigKeys.DEFAULT_MKTXP_SOCKET_TIMEOUT,
            initial_delay_on_failure=MKTXPConfigKeys.DEFAULT_MKTXP_INITIAL_DELAY,
            max_delay_on_failure=MKTXPConfigKeys.DEFAULT_MKTXP_MAX_DELAY,
            delay_inc_div=MKTXPConfigKeys.DEFAULT_MKTXP_INC_DIV,
            bandwidth=False,
            verbose_mode=False,
            bandwidth_test_interval=MKTXPConfigKeys.DEFAULT_MKTXP_BANDWIDTH_TEST_INTERVAL,
            minimal_collect_interval=MKTXPConfigKeys.DEFAULT_MKTXP_MIN_COLLECT_INTERVAL,
            fetch_routers_in_parallel=False,
            max_worker_threads=MKTXPConfigKeys.DEFAULT_MKTXP_MAX_WORKER_THREADS,
            max_scrape_duration=MKTXPConfigKeys.DEFAULT_MKTXP_MAX_SCRAPE_DURATION,
            total_max_scrape_duration=MKTXPConfigKeys.DEFAULT_MKTXP_TOTAL_MAX_SCRAPE_DURATION,
            compact_default_conf_values=False,
            prometheus_headers_deduplication=False,
            persistent_router_connection_pool=True
        )

class MKTXPConfigHandler:
    # two-phase init
    def __init__(self):
        # to be re-inisialised properly in the second phase
        self.system_entry = mockSystemEntry

    # two-phase init, to enable custom config
    def __call__(self, os_config = None):
        self.os_config = os_config if os_config else OSConfig.os_config()
        if not self.os_config:
            sys.exit(1)

        # mktxp user config folder
        if not os.path.exists(self.os_config.mktxp_user_dir_path):
            os.makedirs(self.os_config.mktxp_user_dir_path)

        # if needed, stage the user config data
        self.usr_conf_data_path = os.path.join(
            self.os_config.mktxp_user_dir_path, 'mktxp.conf')
        self.mktxp_conf_path = os.path.join(
            self.os_config.mktxp_user_dir_path, '_mktxp.conf')

        self._create_os_path(self.usr_conf_data_path,
                             'cli/config/mktxp.conf')
        self._create_os_path(self.mktxp_conf_path,
                             'cli/config/_mktxp.conf')

        self.re_compiled = {}

        self._read_from_disk()

        self.default_config_entry_reader = self._default_config_entry_reader()
        self.system_entry = self._system_entry()

    # MKTXP entries
    def registered_entries(self):
        ''' All MKTXP registered entries
        '''
        return (entry_name for entry_name in self.config.keys() if entry_name != MKTXPConfigKeys.DEFAULT_ENTRY_KEY )

    def registered_entry(self, entry_name):
        ''' A specific MKTXP registered entry by name
        '''
        return self.config.get(entry_name)

    def config_entry(self, entry_name):
        ''' Given an entry name, reads and returns the entry info
        '''
        entry_reader = self._config_entry_reader(entry_name)
        return ConfigEntry.MKTXPConfigEntry(**entry_reader) if entry_reader else None

    # Helpers
    def _system_entry(self):
        ''' MKTXP internal config entry
        '''
        _entry_reader = self._system_entry_reader()
        return ConfigEntry.MKTXPSystemEntry(**_entry_reader)

    def _read_from_disk(self):
        ''' (Force-)Read conf data from disk
        '''
        self.config = ConfigObj(self.usr_conf_data_path, indent_type = '    ')
        self.config.preserve_comments = True

        self._config = ConfigObj(self.mktxp_conf_path, indent_type = '    ')
        self._config.preserve_comments = True

    def _create_os_path(self, os_path, resource_path):
        if not os.path.exists(os_path):
            # stage from the conf templates
            ref = importlib.resources.files('mktxp') / resource_path
            with importlib.resources.as_file(ref) as path:
                shutil.copy(path, os_path)

    def _system_entry_reader(self):
        system_entry_reader = {}
        entry_name = MKTXPConfigKeys.MKTXP_CONFIG_ENTRY_NAME
        new_keys = []

        for key in MKTXPConfigKeys.MKTXP_INT_KEYS:
            if self._config[entry_name].get(key):
                system_entry_reader[key] = self._config[entry_name].as_int(key)
            else:
                system_entry_reader[key] = self._default_value_for_key(key)
                if key not in (MKTXPConfigKeys.PORT_KEY):  # Port key has been depricated
                    new_keys.append(key) # read from disk next time

        for key in MKTXPConfigKeys.SYSTEM_BOOLEAN_KEYS_NO.union(MKTXPConfigKeys.SYSTEM_BOOLEAN_KEYS_YES):
            if self._config[entry_name].get(key) is not None:
                system_entry_reader[key] = self._config[entry_name].as_bool(key)
            else:
                system_entry_reader[key] = True if key in MKTXPConfigKeys.SYSTEM_BOOLEAN_KEYS_YES else False
                new_keys.append(key) # read from disk next time

        # listen 
        if self._config[entry_name].get(MKTXPConfigKeys.LISTEN_KEY):
            system_entry_reader[MKTXPConfigKeys.LISTEN_KEY] = self._config[entry_name].get(MKTXPConfigKeys.LISTEN_KEY)
        else:
            system_entry_reader[MKTXPConfigKeys.LISTEN_KEY] = f'0.0.0.0:{system_entry_reader[MKTXPConfigKeys.PORT_KEY]}'
            new_keys.append(MKTXPConfigKeys.LISTEN_KEY) # read from disk next time

        if new_keys:
            self._config[entry_name] = system_entry_reader
            try:
                self._config[entry_name].pop(MKTXPConfigKeys.PORT_KEY, None) # Port key has been depricated
                self._config.write()
                if self._config[entry_name].as_bool(MKTXPConfigKeys.MKTXP_VERBOSE_MODE):
                    print(f'Updated system entry {entry_name} with new system keys {new_keys}')    
            except Exception as exc:
                print(f'Error updating system entry {entry_name} with new system keys {new_keys}: {exc}')
                print('Please update _mktxp.conf to its latest version manually')

        return system_entry_reader

    def _config_entry_reader(self, entry_name):
        config_entry_reader = {}
        compact_config = self._config[MKTXPConfigKeys.MKTXP_CONFIG_ENTRY_NAME].as_bool(MKTXPConfigKeys.MKTXP_COMPACT_CONFIG)
        drop_keys = []

        for key in MKTXPConfigKeys.BOOLEAN_KEYS_NO.union(MKTXPConfigKeys.BOOLEAN_KEYS_YES):
            if self.config[entry_name].get(key) is not None:
                config_entry_reader[key] = self.config[entry_name].as_bool(key)
                if compact_config and config_entry_reader[key] == self.default_config_entry_reader[key]:
                    drop_keys.append(key)
            else:
                config_entry_reader[key] = self.default_config_entry_reader[key]

        for key in MKTXPConfigKeys.STR_KEYS:
            if self.config[entry_name].get(key):
                config_entry_reader[key] = self.config[entry_name].get(key)
                if key is MKTXPConfigKeys.PASSWD_KEY and type(config_entry_reader[key]) is list:
                    config_entry_reader[key] = ','.join(config_entry_reader[key])         

                if compact_config and config_entry_reader[key] == self.default_config_entry_reader[key]:
                    drop_keys.append(key)
            else:
                config_entry_reader[key] = self.default_config_entry_reader[key]

        for key in MKTXPConfigKeys.INT_KEYS:
            if self.config[entry_name].get(key):
                config_entry_reader[key] = self.config[entry_name].as_int(key)
                if compact_config and config_entry_reader[key] == self.default_config_entry_reader[key]:
                    drop_keys.append(key)                
            else:
                config_entry_reader[key] = self.default_config_entry_reader[key]

        # port
        if self.config[entry_name].get(MKTXPConfigKeys.PORT_KEY):
            config_entry_reader[MKTXPConfigKeys.PORT_KEY] = self.config[entry_name].as_int(MKTXPConfigKeys.PORT_KEY)
            if compact_config and config_entry_reader[MKTXPConfigKeys.PORT_KEY] == self.default_config_entry_reader[MKTXPConfigKeys.PORT_KEY]:
                drop_keys.append(MKTXPConfigKeys.PORT_KEY)    
        else:
            config_entry_reader[MKTXPConfigKeys.PORT_KEY] = self.default_config_entry_reader[MKTXPConfigKeys.PORT_KEY]

        # If allowed, compact mktxp.conf entry
        if drop_keys and compact_config:
            for key in drop_keys:
                self.config[entry_name].pop(key, None)
            try:
                self.config.write()
                if self._config[MKTXPConfigKeys.MKTXP_CONFIG_ENTRY_NAME].as_bool(MKTXPConfigKeys.MKTXP_VERBOSE_MODE):
                    print(f'compacted router entry {entry_name} for default values of the feature keys {drop_keys}')                    
            except Exception as exc:
                print(f'Error compacting router entry {entry_name} for default values of feature keys {drop_keys}: {exc}')
                print(f'Error compacting router entry {entry_name} for default values of feature keys {drop_keys}: {exc}')
                print('Please compact mktxp.conf manually')

        return config_entry_reader

    def _default_config_entry_reader(self):
        default_config_entry_reader = {}
        new_keys = []

        if not self.config.get(MKTXPConfigKeys.DEFAULT_ENTRY_KEY):
            self.config[MKTXPConfigKeys.DEFAULT_ENTRY_KEY] = {}

        for key in MKTXPConfigKeys.BOOLEAN_KEYS_NO.union(MKTXPConfigKeys.BOOLEAN_KEYS_YES):
            if self.config[MKTXPConfigKeys.DEFAULT_ENTRY_KEY].get(key) is not None:
                default_config_entry_reader[key] = self.config[MKTXPConfigKeys.DEFAULT_ENTRY_KEY].as_bool(key)
            else:
                default_config_entry_reader[key] = True if key in MKTXPConfigKeys.BOOLEAN_KEYS_YES else False
                new_keys.append(key) # read from disk next time

        for key in MKTXPConfigKeys.STR_KEYS:
            if self.config[MKTXPConfigKeys.DEFAULT_ENTRY_KEY].get(key) is not None:
                default_config_entry_reader[key] = self.config[MKTXPConfigKeys.DEFAULT_ENTRY_KEY].get(key)
            else:
                default_config_entry_reader[key] = self._default_value_for_key(key)
                new_keys.append(key) # read from disk next time

        for key in MKTXPConfigKeys.INT_KEYS:
            if self.config[MKTXPConfigKeys.DEFAULT_ENTRY_KEY].get(key):
                default_config_entry_reader[key] = self.config[MKTXPConfigKeys.DEFAULT_ENTRY_KEY].as_int(key)
            else:
                default_config_entry_reader[key] = self._default_value_for_key(key)
                new_keys.append(key) # read from disk next time

        # port
        if self.config[MKTXPConfigKeys.DEFAULT_ENTRY_KEY].get(MKTXPConfigKeys.PORT_KEY):
            default_config_entry_reader[MKTXPConfigKeys.PORT_KEY] = self.config[MKTXPConfigKeys.DEFAULT_ENTRY_KEY].as_int(MKTXPConfigKeys.PORT_KEY)
        else:
            default_config_entry_reader[MKTXPConfigKeys.PORT_KEY] = self._default_value_for_key(
                MKTXPConfigKeys.SSL_KEY, default_config_entry_reader[MKTXPConfigKeys.SSL_KEY])
            new_keys.append(MKTXPConfigKeys.PORT_KEY) # read from disk next time
        
        if new_keys:
            self.config[MKTXPConfigKeys.DEFAULT_ENTRY_KEY] = default_config_entry_reader
            try:
                self.config.write()
                if self._config[MKTXPConfigKeys.MKTXP_CONFIG_ENTRY_NAME].as_bool(MKTXPConfigKeys.MKTXP_VERBOSE_MODE):
                    print(f'Updated default router entry with new feature keys {new_keys}')                    
            except Exception as exc:
                print(f'Error updating default router entry with new feature keys {new_keys}: {exc}')
                print('Please update mktxp.conf to its latest version manually')

        return default_config_entry_reader

    def _default_value_for_key(self, key, value=None):
        return {
            MKTXPConfigKeys.SSL_KEY: lambda value: MKTXPConfigKeys.DEFAULT_API_SSL_PORT if value else MKTXPConfigKeys.DEFAULT_API_PORT,
            MKTXPConfigKeys.HOST_KEY: lambda _: MKTXPConfigKeys.DEFAULT_HOST_KEY,
            MKTXPConfigKeys.USER_KEY: lambda _: MKTXPConfigKeys.DEFAULT_USER_KEY,
            MKTXPConfigKeys.PASSWD_KEY: lambda _: MKTXPConfigKeys.DEFAULT_PASSWORD_KEY,
            MKTXPConfigKeys.CREDENTIALS_FILE_KEY: lambda _: MKTXPConfigKeys.DEFAULT_CREDENTIALS_FILE_KEY,
            MKTXPConfigKeys.PORT_KEY: lambda _: MKTXPConfigKeys.DEFAULT_MKTXP_PORT,
            MKTXPConfigKeys.SSL_CA_FILE: lambda _: MKTXPConfigKeys.DEFAULT_SSL_CA_FILE,
            MKTXPConfigKeys.FE_REMOTE_DHCP_ENTRY:  lambda _: MKTXPConfigKeys.DEFAULT_FE_REMOTE_DHCP_ENTRY,
            MKTXPConfigKeys.FE_REMOTE_CAPSMAN_ENTRY:  lambda _: MKTXPConfigKeys.DEFAULT_FE_REMOTE_CAPSMAN_ENTRY,
            MKTXPConfigKeys.MKTXP_SOCKET_TIMEOUT: lambda _: MKTXPConfigKeys.DEFAULT_MKTXP_SOCKET_TIMEOUT,
            MKTXPConfigKeys.MKTXP_INITIAL_DELAY: lambda _: MKTXPConfigKeys.DEFAULT_MKTXP_INITIAL_DELAY,
            MKTXPConfigKeys.MKTXP_MAX_DELAY: lambda _: MKTXPConfigKeys.DEFAULT_MKTXP_MAX_DELAY,
            MKTXPConfigKeys.MKTXP_INC_DIV: lambda _: MKTXPConfigKeys.DEFAULT_MKTXP_INC_DIV,
            MKTXPConfigKeys.MKTXP_BANDWIDTH_TEST_INTERVAL: lambda _: MKTXPConfigKeys.DEFAULT_MKTXP_BANDWIDTH_TEST_INTERVAL,
            MKTXPConfigKeys.MKTXP_MIN_COLLECT_INTERVAL: lambda _: MKTXPConfigKeys.DEFAULT_MKTXP_MIN_COLLECT_INTERVAL,
            MKTXPConfigKeys.MKTXP_MAX_WORKER_THREADS: lambda _: MKTXPConfigKeys.DEFAULT_MKTXP_MAX_WORKER_THREADS,
            MKTXPConfigKeys.MKTXP_MAX_SCRAPE_DURATION: lambda _: MKTXPConfigKeys.DEFAULT_MKTXP_MAX_SCRAPE_DURATION,
            MKTXPConfigKeys.MKTXP_TOTAL_MAX_SCRAPE_DURATION: lambda _: MKTXPConfigKeys.DEFAULT_MKTXP_TOTAL_MAX_SCRAPE_DURATION,
        }[key](value)


# Simplest possible Singleton impl
config_handler = MKTXPConfigHandler()



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.datasource.base_ds import BaseDSProcessor


class HealthMetricsDataSource:
    ''' Health Metrics data provider
    '''             
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None, translation_table = None, translate_if_no_value = False):
        if metric_labels is None:
            metric_labels = []                
        try:
            health_records = router_entry.api_connection.router_api().get_resource('/system/health').get()
            for record in health_records:
                if 'name' in record:
                    # Note: The API in RouterOS v7.X+ returns a response like this:
                    # [{'name': 'temperature', 'value': '33', 'type': 'C'}, ...]
                    # To make this work for both v6 and v7 add a <name>:<value> pair in v7
                    # Otherwise it is not possible to get the value by name (e.g. records['voltage'])
                    name = record['name']
                    val = record.get('value', None)
                    record[name] = val

            return BaseDSProcessor.trimmed_records(router_entry, router_records = health_records, metric_labels = metric_labels, 
                                                    translation_table = translation_table, translate_if_no_value = translate_if_no_value)
        except Exception as exc:
            print(f'Error getting system health info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.health_ds import HealthMetricsDataSource
from mktxp.utils.utils import str2bool

class HealthCollector(BaseCollector):
    ''' System Health Metrics collector
    '''    
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.health:
            return

        health_labels = ['voltage', 'temperature', 'phy_temperature', 'cpu_temperature', 'switch_temperature', 
                        'fan1_speed', 'fan2_speed', 'fan3_speed', 'fan4_speed', 'power_consumption', 'board_temperature1', 'board_temperature2',
                        'psu1_voltage', 'psu2_voltage', 'psu1_current', 'psu2_current', 'psu1_state', 'psu2_state', 
                        'poe_out_consumption', 'jack_voltage', '2pin_voltage', 'poe_in_voltage']
        translation_table = {
                'psu1_state': lambda value: '1' if str2bool(value, default = False) else '0',
                'psu2_state': lambda value: '1' if str2bool(value, default = False) else '0'}
        health_records = HealthMetricsDataSource.metric_records(router_entry, metric_labels = health_labels, translation_table = translation_table)   
        if health_records:
            for record in health_records:

                if 'voltage' in record:
                    voltage_metrics = BaseCollector.gauge_collector('system_routerboard_voltage', 'Supplied routerboard voltage', [record, ], 'voltage')
                    yield voltage_metrics

                if 'temperature' in record:
                    temperature_metrics = BaseCollector.gauge_collector('system_routerboard_temperature', 'Routerboard current temperature', [record, ], 'temperature')
                    yield temperature_metrics

                if 'phy_temperature' in record:
                    phy_temperature_metrics = BaseCollector.gauge_collector('system_phy_temperature', 'Current PHY temperature', [record, ], 'phy_temperature')
                    yield phy_temperature_metrics

                if 'cpu_temperature' in record:
                    cpu_temperature_metrics = BaseCollector.gauge_collector('system_cpu_temperature', 'Current CPU temperature', [record, ], 'cpu_temperature')
                    yield cpu_temperature_metrics
                
                if 'switch_temperature' in record:
                    switch_temperature_metrics = BaseCollector.gauge_collector('system_switch_temperature', 'Current switch temperature', [record, ], 'switch_temperature')
                    yield switch_temperature_metrics

                if 'fan1_speed' in record:
                    fan_one_speed_metrics = BaseCollector.gauge_collector('system_fan_one_speed', 'System fan 1 current speed', [record, ], 'fan1_speed')
                    yield fan_one_speed_metrics

                if 'fan2_speed' in record:
                    fan_two_speed_metrics = BaseCollector.gauge_collector('system_fan_two_speed', 'System fan 2 current speed', [record, ], 'fan2_speed')
                    yield fan_two_speed_metrics

                if 'fan3_speed' in record:
                    fan_three_speed_metrics = BaseCollector.gauge_collector('system_fan_three_speed', 'System fan 3 current speed', [record, ], 'fan3_speed')
                    yield fan_three_speed_metrics

                if 'fan4_speed' in record:
                    fan_four_speed_metrics = BaseCollector.gauge_collector('system_fan_four_speed', 'System fan 4 current speed', [record, ], 'fan4_speed')
                    yield fan_four_speed_metrics

                if 'power_consumption' in record:
                    power_consumption_metrics = BaseCollector.gauge_collector('system_power_consumption', 'System Power Consumption', [record, ], 'power_consumption')
                    yield power_consumption_metrics

                if 'board_temperature1' in record:
                    board_temperature1_metrics = BaseCollector.gauge_collector('system_board_temperature1', 'System board temperature 1', [record, ], 'board_temperature1')
                    yield board_temperature1_metrics

                if 'board_temperature2' in record:
                    board_temperature2_metrics = BaseCollector.gauge_collector('system_board_temperature2', 'System board temperature 2', [record, ], 'board_temperature2')
                    yield board_temperature2_metrics

                if 'psu1_voltage' in record:
                    psu1_voltage_metrics = BaseCollector.gauge_collector('system_psu1_voltage', 'System PSU1 voltage', [record, ], 'psu1_voltage')
                    yield psu1_voltage_metrics

                if 'psu2_voltage' in record:
                    psu2_voltage_metrics = BaseCollector.gauge_collector('system_psu2_voltage', 'System PSU2 voltage', [record, ], 'psu2_voltage')
                    yield psu2_voltage_metrics

                if 'psu1_current' in record:
                    psu1_current_metrics = BaseCollector.gauge_collector('system_psu1_current', 'System PSU1 current', [record, ], 'psu1_current')
                    yield psu1_current_metrics

                if 'psu2_current' in record:
                    psu2_current_metrics = BaseCollector.gauge_collector('system_psu2_current', 'System PSU2 current', [record, ], 'psu2_current')
                    yield psu2_current_metrics

                if 'psu1_state' in record:
                    yield BaseCollector.gauge_collector('system_psu1_state', 'System PSU1 state', [record, ], 'psu1_state')

                if 'psu2_state' in record:
                    yield BaseCollector.gauge_collector('system_psu2_state', 'System PSU2 state', [record, ], 'psu2_state')

                if 'poe_out_consumption' in record:
                    poe_out_consumption_metrics = BaseCollector.gauge_collector('system_poe_out_consumption', 'System POE-out consumption', [record, ], 'poe_out_consumption')
                    yield poe_out_consumption_metrics

                if 'jack_voltage' in record:
                    jack_voltage_metrics = BaseCollector.gauge_collector('system_jack_voltage', 'System Jack Voltage', [record, ], 'jack_voltage')
                    yield jack_voltage_metrics

                if '2pin_voltage' in record:
                    pin_voltage_metrics = BaseCollector.gauge_collector('system_2pin_voltage', 'System 2pin Voltage', [record, ], '2pin_voltage')
                    yield pin_voltage_metrics

                if 'poe_in_voltage' in record:
                    poe_in_voltage_metrics = BaseCollector.gauge_collector('system_poe_in_voltage', 'System POE-in Voltage', [record, ], 'poe_in_voltage')
                    yield poe_in_voltage_metrics




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.user_ds import UserMetricsDataSource

class UserCollector(BaseCollector):
    '''Active Users collector'''
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.user:
            return

        user_labels = ['name', 'when', 'address', 'via', 'group']
        user_records = UserMetricsDataSource.metric_records(router_entry, metric_labels=user_labels)        
        if user_records:
            # Auto de-duplicate the records from the scrapes via API silently
            user_metrics = BaseCollector.info_collector('active_users', 'Active Users', user_records, user_labels, verbose_reporting = False)
            yield user_metrics



# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from collections import namedtuple
from mktxp.datasource.base_ds import BaseDSProcessor


class IPConnectionDatasource:
    ''' IP connections data provider
    '''             
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None):
        if metric_labels is None:
            metric_labels = []        
        try:
            res = router_entry.api_connection.router_api().get_resource('/ip/firewall/connection/').call('print', {'count-only': ''})
            # result processing as described at: https://github.com/socialwifi/RouterOS-api/issues/79#issuecomment-2089744809
            cnt_str = res.done_message.get('ret')
            try:
                count = int(cnt_str)
            except (ValueError, TypeError):
                cnt_str = '0'
            records = [{'count': cnt_str}]            
            return BaseDSProcessor.trimmed_records(router_entry, router_records = records, metric_labels = metric_labels)
        except Exception as exc:
            print(f'Error getting IP connection info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None


class IPConnectionStatsDatasource:
    ''' IP connections stats data provider
    '''             
    @staticmethod
    def metric_records(router_entry, *, metric_labels = None, add_router_id = True):
        if metric_labels is None:
            metric_labels = []        
        try:
            # First, check if there are any connections
            count_records = IPConnectionDatasource.metric_records(router_entry)
            if count_records[0].get('count', 0) == '0':
                return []

            connection_records = router_entry.api_connection.router_api().get_resource('/ip/firewall/connection/').call('print', \
                                                                                            {'proplist':'src-address,dst-address,protocol'})
             # calculate number of connections per src-address
            connections_per_src_address = {}
            for connection_record in connection_records:
                address = connection_record['src-address'].split(':')[0]
                destination = f"{connection_record.get('dst-address')}({connection_record.get('protocol')})"

                count, destinations = 0, set()
                if connections_per_src_address.get(address):
                    count, destinations = connections_per_src_address[address]
                count += 1
                destinations.add(destination)
                connections_per_src_address[address] = ConnStatsEntry(count, destinations)

            # compile connections-per-interface records
            records = []
            for key, entry in connections_per_src_address.items():
                record = {'src_address': key, 'connection_count': entry.count, 'dst_addresses': ', '.join(entry.destinations)}
                if add_router_id: 
                    for router_key, router_value in router_entry.router_id.items():
                        record[router_key] = router_value
                records.append(record)
            return records 
        except Exception as exc:
            print(f'Error getting IP connection stats info from router {router_entry.router_name}@{router_entry.config_entry.hostname}: {exc}')
            return None


ConnStatsEntry = namedtuple('ConnStatsEntry', ['count', 'destinations'])


# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.identity_ds import IdentityMetricsDataSource


class IdentityCollector(BaseCollector):
    ''' System Identity Metrics collector
    '''     
    @staticmethod
    def collect(router_entry):
        identity_labels = ['name']
        identity_records = IdentityMetricsDataSource.metric_records(router_entry, metric_labels = identity_labels)                
        if identity_records:
            identity_metrics = BaseCollector.info_collector('system_identity', 'System identity', identity_records, identity_labels)
            yield identity_metrics




# coding=utf8
## Copyright (c) 2020 Arseniy Kuznetsov
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.


from mktxp.cli.config.config import MKTXPConfigKeys
from mktxp.flow.processor.output import BaseOutputProcessor
from mktxp.collector.base_collector import BaseCollector
from mktxp.datasource.capsman_ds import CapsmanCapsMetricsDataSource, CapsmanRegistrationsMetricsDataSource, CapsmanInterfacesDatasource
from mktxp.datasource.wireless_ds import WirelessMetricsDataSource


class CapsmanCollector(BaseCollector):
    ''' CAPsMAN Metrics collector
    '''    
    @staticmethod
    def collect(router_entry):
        if not router_entry.config_entry.capsman:
            return

        remote_caps_labels = ['identity', 'version', 'base_mac', 'board', 'base_mac']
        remote_caps_records = CapsmanCapsMetricsDataSource.metric_records(router_entry, metric_labels = remote_caps_labels)
        if remote_caps_records:
            remote_caps_metrics = BaseCollector.info_collector('capsman_remote_caps', 'CAPsMAN remote caps', remote_caps_records, remote_caps_labels)
            yield remote_caps_metrics

        registration_labels = ['interface', 'ssid', 'mac_address', 'tx_rate', 'rx_rate', 'rx_signal', 'signal', 'uptime', 'bytes']
        registration_records = CapsmanRegistrationsMetricsDataSource.metric_records(router_entry, metric_labels = registration_labels)
        if registration_records:
            # calculate number of registrations per interface
            registration_per_interface = {}
            for registration_record in registration_records:
                registration_per_interface[registration_record['interface']] = registration_per_interface.get(registration_record['interface'], 0) + 1
            # compile registrations-per-interface records
            registration_per_interface_records = [{ MKTXPConfigKeys.ROUTERBOARD_NAME: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_NAME],
                                            MKTXPConfigKeys.ROUTERBOARD_ADDRESS: router_entry.router_id[MKTXPConfigKeys.ROUTERBOARD_ADDRESS],
                                            'interface': key, 'count': value} for key, value in registration_per_interface.items()]
            # yield registrations-per-interface metrics
            registration_per_interface_metrics = BaseCollector.gauge_collector('capsman_registrations_count', 'Number of active registration per CAPsMAN interface', registration_per_interface_records, 'count', ['interface'])
            yield registration_per_interface_metrics

            # the client info metrics
            if router_entry.config_entry.capsman_clients:

                # translate / trim / augment registration records
                for registration_record in registration_records:
                    BaseOutputProcessor.augment_record(router_entry, registration_record)
                    
                tx_byte_metrics = BaseCollector.counter_collector('capsman_clients_tx_bytes', 'Number of sent packet bytes', registration_records, 'tx_bytes', ['dhcp_name', 'mac_address'])
                yield tx_byte_metrics

                rx_byte_metrics = BaseCollector.counter_collector('capsman_clients_rx_bytes', 'Number of received packet bytes', registration_records, 'rx_bytes', ['dhcp_name', 'mac_address'])
                yield rx_byte_metrics

                signal_strength_metrics = BaseCollector.gauge_collector('capsman_clients_signal_strength', 'Client devices signal strength', registration_records, 'rx_signal', ['dhcp_name', 'mac_address'])
                yield signal_strength_metrics

                registration_metrics = BaseCollector.info_collector('capsman_clients_devices', 'Registered client devices info', 
                                        registration_records, ['dhcp_name', 'dhcp_address', 'rx_signal', 'ssid', 'tx_rate', 'rx_rate', 'interface', 'mac_address', 'uptime'])
                yield registration_metrics


        remote_cap_interface_labels = ['name', 'configuration', 'mac_address', 'current_state', 'current_channel', 'current_registered_clients']
        remote_cap_interface_records = CapsmanInterfacesDatasource.metric_records(router_entry, metric_labels = remote_cap_interface_labels)
        if remote_cap_interface_records:
            remote_caps_metrics = BaseCollector.info_collector('capsman_interfaces', 'CAPsMAN interfaces', remote_cap_interface_records, remote_cap_interface_labels)
            yield remote_caps_metrics

